########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations

from .. import token_type as ttype

class Node:
    def __init__(self, literal: str) -> None:
        self.literal = literal
    def __repr__(self) -> str:
        return str()

class Statement(Node):
    def __init__(self, literal: str) -> None:
        super().__init__(literal)

    def __repr__(self) -> str:
        return f"{self.literal}"

class BlockStatement(Statement):
    def __init__(self, token: ttype.Token, statements: list[Statement]) -> None:
        super().__init__(token.string)
        self.token = token
        self.statements = statements

    def __repr__(self) -> str:
        result = "{\n"
        for statement in self.statements:
            result += "    " + repr(statement) + ";\n"
        return result + "}"

class Expression(Node):
    def __init__(self, literal: str) -> None:
        super().__init__(literal)

    def __repr__(self) -> str:
        return f"{self.literal}"

class Program(Node):
    def __init__(self, statements: list[Statement]) -> None:
        self.statements = statements
    
    def __repr__(self) -> str:
        result = ""
        for statement in self.statements:
            result += repr(statement) + "\n"
        result = result[:-1]

        return result

class Identifier(Expression):
    def __init__(self, token: ttype.Token) -> None:
        super().__init__(token.string)
        self.token = token

    def __repr__(self) -> str:
        return f"{self.literal}"

class Number(Expression):
    def __init__(self, token: ttype.Token) -> None:
        super().__init__(token.string)
        self.token = token
        assert isinstance(token.value, float)
        self.value: float = token.value

    def __repr__(self) -> str:
        return self.literal

class String(Expression):
    def __init__(self, token: ttype.Token) -> None:
        super().__init__(token.string)
        self.token = token
        self.string = token.string

    def __repr__(self) -> str:
        return "\"" + self.string + "\""

class Boolean(Expression):
    def __init__(self, token: ttype.Token, value: bool) -> None:
        super().__init__(token.string)
        self.token = token
        self.value = value
    
    def __repr__(self) -> str:
        return self.literal

class Null(Expression):
    def __init__(self, token: ttype.Token) -> None:
        super().__init__(token.string)
        self.token = token

    def __repr__(self) -> str:
        return self.literal

class PrefixExpression(Expression):
    def __init__(self, token: ttype.Token, rightexpr: Expression) -> None:
        super().__init__(token.string)
        self.token = token
        self.operator = self.literal
        self.rightexpr = rightexpr

    def __repr__(self) -> str:
        return f"({self.operator}{self.rightexpr})"

class InfixExpression(Expression):
    def __init__(self, token: ttype.Token, leftexpr: Expression, rightexpr: Expression) -> None:
        super().__init__(token.string)
        self.token = token
        self.leftexpr = leftexpr
        self.operator = self.literal
        self.rightexpr = rightexpr

    def __repr__(self) -> str:
        return f"({self.leftexpr} {self.operator} {self.rightexpr})"

class IfExpression(Expression):
    def __init__(self, token: ttype.Token, condition: Expression, consequence: BlockStatement, alternative: BlockStatement) -> None:
        super().__init__(token.string)
        self.token = token
        self.condition = condition
        self.consequence = consequence
        self.alternative = alternative

    def __repr__(self) -> str:
        result = f"if {self.condition} {self.consequence}{' else '+ repr(self.alternative) if self.alternative.literal != 'if' else ''}"
        return result

class FunctionLiteral(Expression):
    def __init__(self, token: ttype.Token, parameters: list[Identifier], body: BlockStatement) -> None:
        super().__init__(token.string)
        self.token = token
        self.parameters = parameters
        self.body = body
    
    def __repr__(self) -> str:
        result = f"{self.literal}({', '.join([repr(parametre) for parametre in self.parameters])}) {self.body}"

        return result

class ArrayLiteral(Expression):
    def __init__(self, token: ttype.Token, elements: list[Expression]) -> None:
        super().__init__(token.string)
        self.elements = elements

    def __repr__(self) -> str:
        return repr(self.elements)

class HashLiteral(Expression):
    def __init__(self, token: ttype.Token, pairs: dict[Expression, Expression]) -> None:
        super().__init__(token.string)
        self.token = token
        self.pairs = pairs
    
    def __repr__(self) -> str:
        return repr(self.pairs)

class LetStatement(Statement):
    def __init__(self, token: ttype.Token, name: Identifier, value: Expression) -> None:
        super().__init__(token.string)
        self.token = token
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return f"{self.literal} {self.name} = {self.value}"

class IndexExpression(Expression):
    def __init__(self, token: ttype.Token, left: ArrayLiteral | HashLiteral, index: Expression) -> None:
        super().__init__(token.string)
        self.token = token
        self.left = left
        self.index = index

    def __repr__(self) -> str:
        return f"{self.left}[{self.index}]"

class ReassignmentStatement(Statement):
    def __init__(self, token: ttype.Token, index_expr: IndexExpression, value: Expression) -> None:
        super().__init__(token.string)
        self.token = token
        self.index_expr = index_expr
        self.value = value

    def __repr__(self) -> str:
        return f"{self.literal} {self.index_expr} = {self.value}"

class ReturnStatement(Statement):
    def __init__(self, token: ttype.Token, expression: Expression) -> None:
        super().__init__(token.string)
        self.token = token
        self.valuexp = expression

    def __repr__(self) -> str:
        return f"{self.literal} {self.valuexp}"

class CallExpression(Expression):
    def __init__(self, token: ttype.Token, function: FunctionLiteral, arguments: list[Expression]) -> None:
        super().__init__(token.string)
        self.token = token = token
        self.function = function
        self.arguments = arguments
    
    def __repr__(self) -> str:
        result = f"{self.function}({', '.join(repr(argument) for argument in self.arguments)})"
        return result

class ExpressionStatement(Statement):
    def __init__(self, token: ttype.Token, expression: Expression) -> None:
        super().__init__(token.string)
        self.token = token
        self.expression = expression

    def __repr__(self) -> str:
        return f"{self.expression}"



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from .error import (
    Error as Error,
    IllegalCharErr as IllegalCharErr,
    UnexpectedToken as UnexpectedToken,
    NoPrefixFunc as NoPrefixFunc,
    UnknownOperator as UnknownOperator,
    UnknownIdentifier as UnknownIdentifier,
    NotAFunction as NotAFunction,
    UnexpectedEOF as UnexpectedEOF,
    ArgumentError as ArgumentError,
    UnsupportedType as UnsupportedType,
    UnsupportedIndexAccessType as UnsupportedIndexAccessType,
    UnsupportedIndexType as UnsupportedIndexType,
    UnsupporteKeyType as UnsupporteKeyType,
    UnknownNode as UnknownNode,
)



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations
import readline  # noqa: F401
from sys import argv
from typing import Optional, Union

from .astt import Program
from . import error
from .evaluator import evaluate
from .objectt import NULL, Enviroment
from . import parser
from .token_type import TT_ERR, Token
from . import lexer
from . import objectt as obj


def run(fname: Optional[str], source: str, env) -> tuple[Union[Program, list[Token], str], Union[list[error.Error], None]]:
    def check_err(tokens):
        for token in tokens:
            if token.ttype == TT_ERR:
                return True
        return False

    def get_errs(tokens):
        errs = []
        for token in tokens:
            if token.ttype == TT_ERR:
                errs.append(token.value)
        return errs
    if fname is None:
        tokens = lexer.tokenize(source)
    else:
        tokens = lexer.tokenize(source, fname=fname)

    if check_err(tokens):
        return tokens, get_errs(tokens)

    program, errors = parser.parse(tokens)

    if len(errors) > 0:
        return program, errors

    result = str()
    result_obj = evaluate(program, env)
    if isinstance(result_obj, obj.Error):
        result = result_obj.inspect()
        errors.append(result_obj.error)
        return result, errors

    if result_obj != NULL:
        result = result_obj.inspect()

    return result, None

def repl():
    print(
    """    Peelpreter, Copyright (C) 2024 Jeebak Samajdwar

    This program comes with ABSOLUTELY NO WARRANTY; type "warranty" 
    for more details.
    This is free software, and you are welcome to redistribute it
    under certain conditions; type "license" for details or visit 
    https://www.gnu.org/licenses/gpl-3.0.html.\n"""
    )
    print("Welcome! This is the the monkey programming language!")
    env = Enviroment(None)
    while True:
        try:
            source = input(">> ")
        except (EOFError, KeyboardInterrupt):
            print()
            return 

        if source == "license":
            print(
                """    Peelpreter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version."""
            )
            continue
        elif source == "warranty":
            print( 
                """    Peelpreter is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details."""
            )    
            continue        

        result, errs = run(None, source, env)
        if not errs:
            if result == "":
                #TODO throw proper error
                continue
            print(result)
        else:
            for err in errs:
                print(str(err))

def main():
    if len(argv) == 1:
        repl()
    elif len(argv) == 2:
        with open(argv[1], "r") as sourcef:
            env = Enviroment(None)
            result, errors = run(argv[1], sourcef.read(), env)
            if errors is not None:
                result = ""
                for error in errors:
                    result += str(error)
                print(result)
    else:
        print("Invalid command line arguments. Please input only one argument being the file name or no argument for REPL")

if __name__ == "__main__":
    main()



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations
from ..error import Error, IllegalCharErr, UnexpectedEOF
from .. import token_type as ttoken

def tokenize(source: str, fname: str = "stdin") -> list[ttoken.Token]:
    def advance(position: int) -> int:
        return position + 1 if position < len(source) -1 else -1
    def peek(position: int) -> str:
        return source[position + 1] if position + 1 != len(source) else ""
    def handle_num(position: int, current: str, tokens: list[ttoken.Token]) -> int:
        num = current
        index = position
        while num.isdigit():
            index += 1
            if not index < len(source):
                break
            num = source[index]
        if not num == ".":
            tokens.append(ttoken.Token(ttoken.TT_NUM, source[position:index], float(source[position:index]))) 
            return index - 1
        if peek(index).isdigit() and num == ".":
            index += 1
            num = source[index]
            while num.isdigit():
                index += 1
                if not index < len(source):
                    break
                num = source[index]
        tokens.append(ttoken.Token(ttoken.TT_NUM, source[position:index], float(source[position:index])))
        return index - 1

    def handle_str(position: int, current: str, tokens: list[ttoken.Token], line: int, column: int) -> int:
        start_pos = position
        position = advance(position)
        if position == -1:
            add_err(UnexpectedEOF(fname, (line, column)), current, tokens)
            return start_pos
        current = source[position]
        while current != "\"":
            old_pos = position
            position = advance(position)
            if position == -1:
                add_err(UnexpectedEOF(fname, (line, column)), current, tokens)
                return old_pos
            current = source[position]

        tokens.append(ttoken.Token(ttoken.TT_STRING, source[start_pos + 1:position], None))
        return position

    def handle_indentifier(position: int, current: str, tokens: list[ttoken.Token]) -> int:
        index = position
        char = current
        while char.isalpha() or char == "_":
            index += 1
            if not index < len(source):
                break
            char = source[index]
        if source[position:index] in ttoken.keywords:
            tokens.append(ttoken.Token(ttoken.keywords[source[position:index]], source[position:index], None))
            return index - 1
        tokens.append(ttoken.Token(ttoken.TT_IDEN, source[position:index], None))
        return index - 1

    def add_token(ttype: str, current: str, tokens: list[ttoken.Token]) -> None:
        tokens.append(ttoken.Token(ttype, current, None))
    def add_err(err: Error, current, tokens: list[ttoken.Token]) -> None:
        tokens.append(ttoken.Token(ttoken.TT_ERR, current, err))

    tokens: list[ttoken.Token] = []
    line = 0
    column = 0
    position = -1
    current = None

    while True:
        position = advance(position)
        if position == -1:
            break
        column += 1
        current = source[position]

        if current == "=":
            if peek(position) == "=" and peek(position) != "":
                position = advance(position)
                add_token(ttoken.TT_EQ, "==", tokens)
                continue
            add_token(ttoken.TT_ASSIGN, current, tokens)
        elif current == "+":
            add_token(ttoken.TT_PLUS, current, tokens)
        elif current == "-":
            add_token(ttoken.TT_MINUS, current, tokens)
        elif current == "/":
            add_token(ttoken.TT_SLASH, current, tokens)
        elif current == "*":
            add_token(ttoken.TT_ASTERISK, current, tokens)
        elif current == "!":
            if peek(position) == "=":
                position = advance(position)
                add_token(ttoken.TT_NTEQ, "!=", tokens)
                continue
            add_token(ttoken.TT_BANG, current, tokens)
        elif current == "<":
            add_token(ttoken.TT_LT, current, tokens)
        elif current == ">":
            add_token(ttoken.TT_GT, current, tokens)
        elif current == "(":
            add_token(ttoken.TT_LPAREN, current, tokens)
        elif current == ")":
            add_token(ttoken.TT_RPAREN, current, tokens)
        elif current == "{":
            add_token(ttoken.TT_LBRACE, current, tokens)
        elif current == "}":
            add_token(ttoken.TT_RBRACE, current, tokens)
        elif current == "[":
            add_token(ttoken.TT_LBRACKET, current, tokens)
        elif current == "]":
            add_token(ttoken.TT_RBRACKET, current, tokens)
        elif current == ",":
            add_token(ttoken.TT_COMMA, current, tokens)
        elif current == ":":
            add_token(ttoken.TT_COLON, current, tokens)
        elif current == " ":
            continue
        elif current == "\t":
            continue
        elif current == "\n":
            column = 1
            line += 1
        elif current == ";":
            add_token(ttoken.TT_SEMICOLON, current, tokens)
        elif current == "\"":
            position = handle_str(position, current, tokens, line, column)
        elif current.isalpha():
            position = handle_indentifier(position, current, tokens)
        elif current.isdigit():
            position = handle_num(position, current, tokens)
        else:
            add_err(IllegalCharErr(fname, current, (line, column)), current, tokens)
                    
    add_token(ttoken.TT_EOF, "", tokens)

    return tokens



########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..token_type import Token
    from ..objectt import Object

class Error:
    def __init__(self, fname: str, errtype: str, info: str, location: tuple[int, int]) -> None:
        self.fname = fname
        self.errtype = errtype
        self.info = info
        self.location = location

    def __str__(self) -> str:
        return f"File <{self.fname}> at line {self.location[0]} and column {self.location[1]},\n{self.errtype}: {self.info}"

class IllegalCharErr(Error):
    def __init__(self, fname: str, char: str, location: tuple[int, int]) -> None:
        super().__init__(fname, "IllegalCharErr", f"illegal charecter '{char}'", location)

class UnexpectedToken(Error):
    def __init__(self, fname: str, expected_token: str, got_token: str, location: tuple[int, int]) -> None:
        super().__init__(fname, "UnexpectedToken", f"expected token {expected_token}, got {got_token}", location)

class NoPrefixFunc(Error):
    def __init__(self, fname: str, token: "Token", location: tuple[int, int]) -> None:
        super().__init__(fname, "NoPrefixFunc", f"no prefix function for parsing {token.string} found", location)

class UnknownOperator(Error):
    def __init__(self, fname: str, operator: str, left_type: str | None, right_type: str, location: tuple[int, int]) -> None:
        if left_type is not None:
            super().__init__(fname, "UnknownOperator", f"unknown operator: {operator} between {left_type} and {right_type}", location)
        else:
            super().__init__(fname, "UnknownOperator", f"unknown operator: {operator} for {right_type}", location)

class UnknownIdentifier(Error):
    def __init__(self, fname: str, name: str, location: tuple[int, int]) -> None:
        super().__init__(fname, "UnknownIdentifier", f"unknown identifier '{name}'", location)

class NotAFunction(Error):
    def __init__(self, fname: str, obj: "Object", location: tuple[int, int]) -> None:
        super().__init__(fname, "NotAFunction", f"{obj.type()} is not a function", location)

class UnexpectedEOF(Error):
    def __init__(self, fname: str, location: tuple[int, int]) -> None:
        super().__init__(fname, "UnexpectedEOF", "sudden unexpected EOF token encountered", location)

class ArgumentError(Error):
    def __init__(self, fname: str, expected: int, got: int, func: str, location: tuple[int, int]) -> None:
        super().__init__(fname, "ArgumentError", f"wrong number of arguments, expected {expected} but got {got} for function {func}", location)

class UnsupportedType(Error):
    def __init__(self, fname: str, got: "Object", func: str, location: tuple[int, int]) -> None:
        super().__init__(fname, "UnsupportedType", f"argument of type {got.type()} not supported for function {func}", location)

class UnsupportedIndexAccessType(Error):
    def __init__(self, fname: str, type: "Object", location: tuple[int, int]) -> None:
        super().__init__(fname, "UnsupportedIndexAccessType", f"{type.type()} not supported for index access expressions", location)

class UnsupportedIndexType(Error):
    def __init__(self, fname: str, got: "Object", location: tuple[int, int]) -> None:
        super().__init__(fname, "UnsupportedIndexType", f"index of type {got.type()} is not supported for array", location)

class UnsupporteKeyType(Error):
    def __init__(self, fname: str, type: "Object", location: tuple[int, int]) -> None:
        super().__init__(fname, "UnsupportedKeyType", f"unsupported type {type.type()} as key", location)

class UnknownNode(Error):
    def __init__(self, fname: str, location: tuple[int, int]) -> None:
        super().__init__(fname, "UnknownNode", "unknown ast node encountered", location)
